#ifdef HAVE_LIBGIF
#include <stdlib.h>
#include <string.h>
#include <gif_lib.h>
#include "imgio.h"


/* Last error message pointer */
extern const char *imgio_last_load_error;
extern const char *imgio_last_write_error;


/* GIF Library Version */
void ImgGIFVersion(int *major, int *minor, int *release); 


/* GIF Reading */
int ImgGIFReadRGBA(
	const char *filename,
	int *width_rtn, int *height_rtn, int *bpp_rtn, int *bpl_rtn,
	u_int8_t **rgba_rtn,
	u_int8_t *bg_color,	/* 4 bytes in RGBA format, will be modified */
	int *x_rtn, int *y_rtn,
	int *base_width_rtn, int *base_height_rtn,
	char **creator_rtn, char **title_rtn,
	char **author_rtn, char **comments_rtn,
	u_int8_t def_alpha_value,
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
);


/* GIF Writing */
static void ImgGIFSplitRGBA(
	const u_int8_t *rgba, int npixels,
	u_int8_t **red_buf,
	u_int8_t **green_buf,
	u_int8_t **blue_buf
);

static int ImgWriteGIFFileRGBAToColor(
	const char *filename,
	int width, int height,
	int bpl, int bpp,       /* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes in RGBA format */  
	const char *comments,   
	int interlaced,         /* 0 or 1 */
	int transparency,       /* 0 or 1 */
	void *client_data,      
	int (*progress_cb)(
		void *,         /* Client data */
		int, int,       /* Progress current & end values */
		int, int,       /* Width & height */
		int, int,       /* Bytes per line & bytes per pixel */  
		u_int8_t *      /* RGBA data */
	),
	int *user_aborted
);
static int ImgWriteGIFFileRGBAToGreyscale(
	const char *filename,
	int width, int height,
	int bpl, int bpp,	/* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes in RGBA format */
	const char *comments,
	int interlaced,		/* 0 or 1 */
	int transparency,	/* 0 or 1 */
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
);
static int ImgWriteGIFFileRGBAToBW(
	const char *filename,
	int width, int height,
	int bpl, int bpp,	/* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes in RGBA format */
	const char *comments,
	int interlaced,		/* 0 or 1 */
	int transparency,	/* 0 or 1 */
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
);
int ImgWriteGIFFileRGBA(
	const char *filename,
	int width, int height,
	int bpl, int bpp,       /* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes in RGBA format */
	const char *comments,
	int interlaced,         /* 0 or 1 */
	int format,             /* 0 = B&W
				 * 1 = Greyscale
				 * 2 = Color */
	int transparency,	/* 0 or 1 */
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	) 
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


static int	gif_interlace_offset[] = { 0, 4, 2, 1 },
		gif_interlace_jumps[] = { 8, 8, 4, 2 };


/*
 *	Gets the GIF library's version.
 */
void ImgGIFVersion(int *major, int *minor, int *release) 
{
	const char *ver_str = GIF_LIB_VERSION, *s;

	s = strstr(ver_str, "Version");
	while((*s != '\0') && (*s != ' '))
	    s++;
	while(*s == ' ')
	    s++;
	if(major != NULL)
	    *major = ATOI(s);

	while((*s != '\0') && (*s != '.'))
	    s++;
	while(*s == '.')
	    s++;
	if(minor != NULL)
	    *minor = ATOI(s);

	while((*s != '\0') && (*s != '.'))
	    s++;
	while(*s == '.')
	    s++;      
	if(release != NULL)
	    *release = ATOI(s);
}


/*
 *	Reads the GIF file.
 *
 *	Return values:
 *
 *	0	Success
 *	-1	General error
 *	-2	Bad value (invalid format/not a gif file)
 *	-3	Systems error
 *	-4	User abort
 */
int ImgGIFReadRGBA(
	const char *filename,
	int *width_rtn, int *height_rtn, int *bpp_rtn, int *bpl_rtn,
	u_int8_t **rgba_rtn,
	u_int8_t *bg_color,     /* 4 bytes in RGBA format, will be modified */
	int *x_rtn, int *y_rtn,
	int *base_width_rtn, int *base_height_rtn,
	char **creator_rtn, char **title_rtn,
	char **author_rtn, char **comments_rtn,
	u_int8_t def_alpha_value,
	void *client_data,
	int (*progress_cb)(
		void *,         /* Client data */
		int, int,       /* Progress current & end values */
		int, int,       /* Width & height */
		int, int,       /* Bytes per line & bytes per pixel */
		u_int8_t *      /* RGBA data */
	), 
	int *user_aborted
)
{
	int i, bpl, buf_len, scr_width, scr_height;
	char *comment = NULL;
	GifFileType *ft;
	GifRowType *buf;
	GifRecordType rt;
	int bg_color_num, transparent_color_num;
	u_int8_t *tar_buf;
	int tar_bpp, tar_bpl;


	/* Reset last load error */
	imgio_last_load_error = NULL;

	/* Skip if user aborted */
	if(*user_aborted)
	{
	    imgio_last_load_error = "User aborted operation";
	    return(-4);
	}

	/* Open GIF file */
	ft = DGifOpenFileName(filename);
	if(ft == NULL)
	{
	    imgio_last_load_error = "Unable to open file for reading";
	    return(-1);
	}

	/* Get GIF screen values */
	scr_width = ft->SWidth;
	scr_height = ft->SHeight;
	bg_color_num = ft->SBackGroundColor;
	transparent_color_num = -1;	/* Initially not set */
#if 0
printf(
 "Total Colors: %i  BG Color: #%i  Res: %i\n",
 ft->SColorMap->ColorCount,
 ft->SBackGroundColor,
 ft->SColorResolution
);
#endif

	/* Get GIF background color */
	if((ft->SColorMap != NULL) && (bg_color != NULL))
	{
	    const ColorMapObject *colormap = ft->SColorMap;
	    const GifColorType *entry = &colormap->Colors[bg_color_num];
	    bg_color[0] = (u_int8_t)entry->Red;
	    bg_color[1] = (u_int8_t)entry->Green;
	    bg_color[2] = (u_int8_t)entry->Blue;
	    bg_color[3] = 0xff;
	}

	/* Allocate target RGBA buffer to the size of the GIF screen */
	tar_bpp = 4;
	tar_bpl = scr_width * tar_bpp;
	tar_buf = (u_int8_t *)calloc(
	    tar_bpl * scr_height, sizeof(u_int8_t)
	);
	if(tar_buf == NULL)
	{
	    DGifCloseFile(ft);
	    imgio_last_load_error = "Memory allocation error";
	    return(-3);
	}

	/* Report initial progress */
	if(progress_cb != NULL)
	{
	    if(!progress_cb(  
		client_data,  
		0, scr_height,
		scr_width, scr_height,
		tar_bpl, tar_bpp,
		tar_buf
	    ))
		*user_aborted = 1;
	}

	/* Allocate screen buffer for reading the images in the GIF
	 * file
	 *
	 * This is an array of GifRowType *
	 */
	buf_len = scr_height * sizeof(GifRowType *);
	buf = (buf_len > 0) ? (GifRowType *)malloc(buf_len) : NULL;
	if(buf == NULL)
	{
	    free(tar_buf);
	    DGifCloseFile(ft);
	    imgio_last_load_error = "Memory allocation error";
	    return(-3);
	}

	/* Bytes per line */
	bpl = scr_width * sizeof(GifPixelType);

	/* Allocate first row and set its color to the background */
	buf[0] = (GifRowType)malloc(bpl);
	for(i = 0; i < scr_width; i++)
	    buf[0][i] = bg_color_num;

	/* Allocate subsequent rows and set their colors to the
	 * background too
	 */
	for(i = 1; i < scr_height; i++)
	{
	    buf[i] = (GifRowType)malloc(bpl);
	    memcpy(buf[i], buf[0], bpl);
	}


	/* Begin reading the GIF file and load the images
	 *
	 * Each image will be rendered on to the target RGBA image
	 * buffer as it is loaded
	 */
	do
	{
	    if(*user_aborted)
		break;

	    if(DGifGetRecordType(ft, &rt) == GIF_ERROR)
		break;

	    switch(rt)
	    {
	      case UNDEFINED_RECORD_TYPE:
		break;

	      case SCREEN_DESC_RECORD_TYPE:
		break;

	      case IMAGE_DESC_RECORD_TYPE:
		if(DGifGetImageDesc(ft) == GIF_OK)
		{
		    GifImageDesc *img = &ft->Image;
		    const ColorMapObject *colormap = (img->ColorMap != NULL) ?
			img->ColorMap : ft->SColorMap;
		    const GifColorType *entry;
		    int	x, y, j, c_num,
			top = img->Top,
			left = img->Left,
			width = img->Width,
			height = img->Height;
		    GifRowType row;
		    u_int8_t *tar_ptr;

		    /* Do we have a colormap (local image or system)? */
		    if((colormap != NULL) ? (colormap->ColorCount <= 0) : TRUE)
			break;

		    /* Load the image to the source buffer */

		    /* Interlaced? */
		    if(img->Interlace)
		    {
			/* Interlace needs 4 passes on the image */
			for(i = 0; i < 4; i++)
			{
			    for(j = top + gif_interlace_offset[i];
				j < (top + height);
				j += gif_interlace_jumps[i]
			    )
			    {  
				if(DGifGetLine(
				    ft,
				    (GifPixelType *)&buf[j][left],
				    width
				) == GIF_ERROR)
				    break;

				/* Report progress */
				if(progress_cb != NULL)
				{
				    if(!progress_cb(
					client_data,
					i, 2 * 4,
					scr_width, scr_height,        
					tar_bpl, tar_bpp,
					tar_buf           
				    ))
				    {
					*user_aborted = 1;
					break;
				    }
				}
			    }
			}    
		    }
		    else
		    {
			/* Read each line sequentially */
			for(i = 0; i < height; i++)
			{
			    if(DGifGetLine(
				ft,
				(GifPixelType *)&buf[top++][left],
				width
			    ) == GIF_ERROR)
				break;

			    /* Report progress */
			    if(progress_cb != NULL)
			    {
				if(!progress_cb(
				    client_data,
				    i, 2 * height,
				    scr_width, scr_height,
				    tar_bpl, tar_bpp,     
				    tar_buf
				))
				{ 
				    *user_aborted = 1;
				    break;
				}
			    }
			}
		    }

		    /* Copy loaded source image data to target buffer */
		    for(y = 0; y < scr_height; y++)
		    {
			row = buf[y];
			tar_ptr = tar_buf + (y * tar_bpl);
			for(x = 0; x < scr_width; x++)
			{
			    /* Current pixel value is colormap index */
			    c_num = (int)row[x];

			    /* Not transparent? */
			    if(c_num != transparent_color_num)
			    {
				entry = &colormap->Colors[c_num];
				*tar_ptr++ = (u_int8_t)entry->Red;
				*tar_ptr++ = (u_int8_t)entry->Green;
				*tar_ptr++ = (u_int8_t)entry->Blue;
				*tar_ptr++ = def_alpha_value;
			    }
			    else
			    {
				tar_ptr += tar_bpp;
			    }
			}

			/* Report progress */
			if(progress_cb != NULL)
			{
			    if(!progress_cb(
				client_data,
				scr_height + y, 2 * scr_height,
				scr_width, scr_height,
				tar_bpl, tar_bpp,
				tar_buf
			    ))
			    {
				*user_aborted = 1;
				break;
			    }
			}
		    }
		}
		break;

	      case EXTENSION_RECORD_TYPE:
		if(1)
		{
		    int code;
		    GifByteType *ext, ext_len;

		    /* Get first extension */
		    if(DGifGetExtension(ft, &code, &ext) == GIF_OK)
		    {
			/* Iterate through extensions */
			while(ext != NULL)
			{
			    /* Get this extension's length */
			    ext_len = ext[0];

			    /* Handle by extension code */
			    switch(code)
			    {
			      case COMMENT_EXT_FUNC_CODE:	/* Comment */
				ext[ext_len + 1] = '\0';
				if(comment != NULL)
				{
				    comment = (char *)realloc(
					comment,
					STRLEN(comment) + ext_len + 1
				    );
				}
				else
				{
				    comment = (char *)malloc(ext_len + 1);
				    *comment = '\0';
				}
				strcat(comment, ext + 1);
				break;

			      case GRAPHICS_EXT_FUNC_CODE:	/* Graphics Control */
				/* Format:
				 *
				 * <len>
				 * <flags> <delay_time> <delay_time_cnt>
				 * <transparent_color>
				 */
				if(ext_len >= 4)
				{
				    unsigned int flags = (unsigned int)ext[1];

				    /* Transparent color set? */
				    if(flags & (1 << 0))       
					transparent_color_num = (int)ext[4];
				}
				break;

			      case PLAINTEXT_EXT_FUNC_CODE:
			      case APPLICATION_EXT_FUNC_CODE:
				break;
			    }

			    /* Get next extension */
			    if(DGifGetExtensionNext(ft, &ext) == GIF_ERROR)
				break;
			}
		    }
		}
		break;

	    case TERMINATE_RECORD_TYPE:
		break;
	    }
	}
	while(rt != TERMINATE_RECORD_TYPE);


	/* Close GIF file */
	DGifCloseFile(ft);

	/* Delete screen buffer */
	for(i = 0; i < scr_height; i++)
	    free(buf[i]);
	free(buf);

	/* Update returns */
	if(width_rtn != NULL)
	    *width_rtn = scr_width;
	if(height_rtn != NULL)
	    *height_rtn = scr_height;
	if(bpl_rtn != NULL)
	    *bpl_rtn = tar_bpl;
	if(bpp_rtn != NULL)
	    *bpp_rtn = tar_bpp;
	if(rgba_rtn != NULL)
	{
	    free(*rgba_rtn);
	    *rgba_rtn = tar_buf;
	    tar_buf = NULL;
	}
	if(base_width_rtn != NULL)
	    *base_width_rtn = scr_width;
	if(base_height_rtn != NULL)
	    *base_height_rtn = scr_height;
	if(comments_rtn != NULL)
	{
	    free(*comments_rtn);
	    *comments_rtn = comment;
	    comment = NULL;
	}

	/* Report final progress */
	if((progress_cb != NULL) && !(*user_aborted))
	{
	    if(!progress_cb(
		client_data,
		scr_height, scr_height,
		scr_width, scr_height,
		tar_bpl, tar_bpp,     
		tar_buf          
	    ))
		*user_aborted = 1;
	}

	/* Delete values that were not transfered to the returns */
	free(tar_buf);
	free(comment);

	if(*user_aborted)
	{
	    imgio_last_load_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(0);
}


/*
 *	Converts the specified RGBA image data to separate Red, Green,
 *	and Blue channel buffers.
 *
 *	Inputs assumed valid.
 */
static void ImgGIFSplitRGBA(
	const u_int8_t *rgba, int npixels,
	u_int8_t **red_buf,
	u_int8_t **green_buf,
	u_int8_t **blue_buf
)
{
	*red_buf = (u_int8_t *)malloc(npixels);
	*green_buf = (u_int8_t *)malloc(npixels);
	*blue_buf = (u_int8_t *)malloc(npixels);
	if((*red_buf != NULL) && (*green_buf != NULL) &&
	   (*blue_buf != NULL)
	)
	{
	    const u_int8_t	*src = rgba,
				*src_end = src + (npixels * 4);

	    u_int8_t	*r_ptr = *red_buf,
			*g_ptr = *green_buf,
			*b_ptr = *blue_buf;

	    while(src < src_end)
	    {
		*r_ptr++ = *src++;
		*g_ptr++ = *src++;
		*b_ptr++ = *src++;
		src++;
	    }
	}
}


/*
 *	Writes the specified image data to a GIF file in color.
 */
static int ImgWriteGIFFileRGBAToColor(
	const char *filename,
	int width, int height,
	int bpl, int bpp,	/* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes in RGBA format */
	const char *comments,
	int interlaced,		/* 0 or 1 */
	int transparency,	/* 0 or 1 */
	void *client_data,
	int (*progress_cb)(
		void *,		/* Client data */
		int, int,	/* Progress current & end values */
		int, int,	/* Width & height */
		int, int,	/* Bytes per line & bytes per pixel */
		u_int8_t *	/* RGBA data */
	),
	int *user_aborted
)
{
	int	color_resolution = 8,
		colormap_ncolors_set = 256,
		bg_color_num = 0,
		transparent_color_num = -1;
	ColorMapObject *colormap = NULL;
	u_int8_t	*red_channel = NULL,
			*green_channel = NULL,
			*blue_channel = NULL;
	GifByteType *buf = NULL;

#define FREE_ALL	{	\
 free(red_channel);		\
 red_channel = NULL;		\
 free(green_channel);		\
 green_channel = NULL;		\
 free(blue_channel);		\
 blue_channel = NULL;		\
				\
 free(buf);			\
 buf = NULL;			\
				\
 if(colormap != NULL) {		\
  free(colormap->Colors);	\
  free(colormap);		\
  colormap = NULL;		\
 }				\
}

	/* Split given RGBA data into separate red, green, and blue
	 * channel buffers
	 */
	ImgGIFSplitRGBA(
	    rgba, width * height,
	    &red_channel, &green_channel, &blue_channel
	);
	if((red_channel == NULL) || (green_channel == NULL) ||
	   (blue_channel == NULL)
	)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}

	/* Reduce number of colors for initial colormap creation to
	 * make room for background and transparent colors
	 */
	if(bg_color != NULL)
	    colormap_ncolors_set--;
	if(transparency)
	    colormap_ncolors_set--;
	if(colormap_ncolors_set < 1)
	    colormap_ncolors_set = 1;

	/* Create colormap */
	colormap = (ColorMapObject *)malloc(sizeof(ColorMapObject));
	if(colormap == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}
	colormap->ColorCount = colormap_ncolors_set;
	colormap->BitsPerPixel = 8;
	colormap->Colors = (GifColorType *)calloc(
	    colormap->ColorCount, sizeof(GifColorType)
	);
	if(colormap->Colors == NULL)
	{
	    colormap->ColorCount = 0;
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}

	/* Create buffer for rendering the output GIF image */
	buf = (GifByteType *)malloc(width * height * sizeof(GifByteType));
	if(buf == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}

	/* Begin converting given image data to the GIF output buffer */
	if(!(*user_aborted))
	{
	    GifFileType *ft;

	    /* Copy red, green, and blue channels to the GIF output
	     * buffer and set the colormap
	     */
	    QuantizeBuffer(
		width, height,
		&colormap_ncolors_set,		/* Colors that have been
						 * set on the colormap
						 * (will be updated by this
						 * call) */
		(GifByteType *)red_channel,
		(GifByteType *)green_channel,
		(GifByteType *)blue_channel,
		buf,
		colormap->Colors
	    );
	    if(colormap_ncolors_set >= colormap->ColorCount)
		colormap_ncolors_set = colormap->ColorCount - 1;
	    if(colormap_ncolors_set < 0)
		colormap_ncolors_set = 0;

	    /* Delete red, green, and blue channels since they are no
	     * longer needed
	     */
	    free(red_channel);
	    red_channel = NULL;
	    free(green_channel);
	    green_channel = NULL;
	    free(blue_channel);
	    blue_channel = NULL;

	    /* Reallocate colormap to full 256 after quantize since
	     * we may have a smaller colormap used for quantize needed
	     * to make room for the background and transparent colors
	     */
	    if(colormap->ColorCount != 256)
	    {
		colormap->ColorCount = 256;
		colormap->Colors = (GifColorType *)realloc(
		    colormap->Colors,
		    colormap->ColorCount * sizeof(GifColorType)
		);
		if(colormap->Colors == NULL)
		{
		    colormap->ColorCount = 0;
		    FREE_ALL
		    imgio_last_write_error = "Memory allocation error";
		    return(-3);
		}
	    }

	    /* Add background color to colormap? */
	    if(bg_color != NULL)
	    {
		int i; 
		GifColorType *c;
		GifByteType *buf_ptr, *buf_end;

		/* Increase ncolors if the colormap has more
		 * allocated colors
		 */
		if(colormap_ncolors_set < colormap->ColorCount)
		    colormap_ncolors_set++;

		/* Shift colors on colormap to accomidate the
		 * background color
		 */
		for(i = colormap_ncolors_set - 1; i > bg_color_num; i--)
		    memcpy(
			&colormap->Colors[i],
			&colormap->Colors[i - 1],
			sizeof(GifColorType)
		    );

		/* Shift color values on output buffer */
		for(buf_ptr = buf,
		    buf_end = buf_ptr + (width * height);
		    buf_ptr < buf_end;
		    buf_ptr++
		)
		{
		    if(*buf_ptr >= (GifByteType)bg_color_num)
			*buf_ptr = (GifByteType)MIN(
			    (int)(*buf_ptr) + 1,
			    0xff
			);
		}

		/* Set background color's value */
		c = &colormap->Colors[bg_color_num];
		c->Red = (GifByteType)bg_color[0];
		c->Green = (GifByteType)bg_color[1];
		c->Blue = (GifByteType)bg_color[2];
	    }

	    /* Apply transparency to the output buffer
	     *
	     * Set all pixels found to be transparent on the source
	     * RGBA data to the transparent color number (last color
	     * index)
	     */
	    if(transparency)
	    {
		int	tar_bpp = sizeof(GifByteType),
			tar_bpl = width * tar_bpp,
			src_bpp = bpp,
			src_bpl = bpl;
		GifByteType *tar_line, *tar_ptr;
		const u_int8_t *src_line, *src_end, *src_line_end, *src_ptr;
		GifColorType *c;

		/* Increase ncolors if the colormap has more 
		 * allocated colors    
		 */
		if(colormap_ncolors_set < colormap->ColorCount)
		    colormap_ncolors_set++;

		/* Set transparent color as the highest set color */
		transparent_color_num = colormap_ncolors_set - 1;

		/* Check for transparent pixels on the source image
		 * buffer and set the corresponding pixel on the
		 * target image buffer to reffer to the new transparent
		 * color
		 */
		for(tar_line = buf,
		    src_line = rgba,
		    src_end = src_line + (bpl * height);
		    src_line < src_end;
		    tar_line += tar_bpl,
		    src_line += src_bpl
		)
		{
		    tar_ptr = tar_line,
		    src_ptr = src_line,
		    src_line_end = src_ptr + src_bpl;

		    while(src_ptr < src_line_end)
		    {
			if(src_ptr[3] < 0x80)
			    *tar_ptr = (GifByteType)transparent_color_num;
			tar_ptr += tar_bpp;
			src_ptr += src_bpp;
		    }
		}

		/* Set transparent color's value */
		c = &colormap->Colors[transparent_color_num];
		c->Red = (GifByteType)0xff;
		c->Green = (GifByteType)0xff;
		c->Blue = (GifByteType)0xff;
	    }

	    /* At this point the GIF output buffer and the colormap are
	     * now ready to be written to the GIF file
	     */

	    /* Open GIF file for writing */
	    ft = EGifOpenFileName(filename, 0);
	    if(ft == NULL)
	    {
		FREE_ALL
		imgio_last_write_error = "Unable to open file for writing";
		return(-1);
	    }

	    /* Write screen descriptor */
	    EGifPutScreenDesc(
		ft,
		width, height,		/* Screen width & height */
		color_resolution,	/* Color resolution */
		bg_color_num,		/* Background color number */
		colormap		/* Screen colormap */
	    );

	    /* Write graphic control */
	    if(transparency && (transparent_color_num > -1))
	    {
		u_int8_t buf[] = {
		    0x01,	/* flags: transparency color flag on */
		    0x00,	/* delay time */
		    0x00,	/* delay time cntd */
		    (u_int8_t)transparent_color_num,
		};
		EGifPutExtension(
		    ft, GRAPHICS_EXT_FUNC_CODE,
		    sizeof(buf), buf
		);
	    }

	    /* Write image */
	    if(!(*user_aborted))
	    {
		/* Write image descriptor */
		EGifPutImageDesc(
		    ft,
		    0, 0,		/* Top & left */
		    width, height,	/* Width & height */
		    interlaced,		/* Interlace */
		    NULL		/* Local image colormap */
		);

		/* Write image data */
		if(interlaced)
		{
		    int i, y;

		    /* Interlace needs 4 passes on the image */
		    for(i = 0; i < 4; i++)
		    {
			for(y = gif_interlace_offset[i];
			    y < height;
			    y += gif_interlace_jumps[i]
			)
			{
			    EGifPutLine(
				ft,
				(GifPixelType *)(buf + (y * width)),
				width
			    );
			}
			/* Report progress */
			if(progress_cb != NULL)
			{
			    if(!progress_cb(
				client_data,
				i, 4,
				width, height,
				bpl, bpp,     
				rgba
			    ))
			    {
				*user_aborted = 1;
				break;            
			    } 
			}    
		    }
		}
		else
		{
		    int y;
		    GifByteType *buf_ptr = buf;
		    for(y = 0; y < height; y++)
		    {
		        EGifPutLine(ft, (GifPixelType *)buf_ptr, width);
		        buf_ptr += width;

		        /* Report progress */
		        if(progress_cb != NULL)
		        {
			    if(!progress_cb(
			        client_data,
			        y, height,
			        width, height,
			        bpl, bpp,
			        rgba
			    ))
			    {
			        *user_aborted = 1;
			        break;
			    }
			}
		    }
		}
	    }

	    /* Write comments */
	    if(!STRISEMPTY(comments))
		EGifPutExtension(
		    ft, COMMENT_EXT_FUNC_CODE,
		    STRLEN(comments) + 1, comments
		);

	    /* Close GIF file */
	    EGifCloseFile(ft);
	}

	FREE_ALL

	if(*user_aborted)
	{
	    imgio_last_write_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(0);

#undef FREE_ALL
}

/*
 *	Writes the specified image data to a GIF file in greyscale.
 */
static int ImgWriteGIFFileRGBAToGreyscale(
	const char *filename,
	int width, int height,
	int bpl, int bpp,       /* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,       /* 4 bytes in RGBA format */
	const char *comments,
	int interlaced,         /* 0 or 1 */
	int transparency,       /* 0 or 1 */
	void *client_data,
	int (*progress_cb)(
		void *,         /* Client data */
		int, int,       /* Progress current & end values */
		int, int,       /* Width & height */
		int, int,       /* Bytes per line & bytes per pixel */
		u_int8_t *      /* RGBA data */
	),
	int *user_aborted
)
{
	int	color_resolution = 8,
		colormap_ncolors_set = 256,
		bg_color_num = 0,
		transparent_color_num = -1;
	ColorMapObject *colormap = NULL;
	GifByteType *buf = NULL;

#define FREE_ALL	{	\
 free(buf);			\
 buf = NULL;			\
				\
 if(colormap != NULL) {		\
  free(colormap->Colors);	\
  free(colormap);		\
  colormap = NULL;		\
 }				\
}

	/* Create colormap */   
	colormap = (ColorMapObject *)malloc(sizeof(ColorMapObject));
	colormap->ColorCount = colormap_ncolors_set;
	colormap->BitsPerPixel = 8;
	colormap->Colors = (GifColorType *)calloc(
	    colormap->ColorCount, sizeof(GifColorType)
	);
	/* Generate greyscale colors */
	if(colormap->Colors != NULL)
	{
	    if(transparency)
	    {
		/* Generate 255 colors from black to white, leaving
		 * color 255 for the transparent pixel
		 */
		int i, m = colormap->ColorCount - 1, n = MAX(m - 1, 1);
		GifColorType *c;
		for(i = 0; i < m; i++)
		{
		    c = &colormap->Colors[i];
		    c->Red = c->Green = c->Blue =
			(GifByteType)(i * 0xff / n);
		}

		/* Set transparent color */
		transparent_color_num = m;
		c = &colormap->Colors[transparent_color_num];
		c->Red = c->Green = c->Blue = 0xff;
	    }
	    else
	    {
		/* Generate 256 colors from black to white */      
		int i, m = colormap->ColorCount, n = MAX(m - 1, 1);
		GifColorType *c;
		for(i = 0; i < m; i++)
		{
		    c = &colormap->Colors[i];
		    c->Red = c->Green = c->Blue =
			(GifByteType)(i * 0xff / n);
		}
	    }
	}

	/* Create buffer for rendering the output GIF image */
	buf = (GifByteType *)malloc(width * height * sizeof(GifByteType));
	if(buf == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}

	/* Copy/convert RGBA data to greyscale output buffer */
	if(transparency && (transparent_color_num > -1))
	{
	    GifByteType	*tar_ptr = buf,
			*tar_end = tar_ptr + (width * height);
	    const u_int8_t *src_ptr = rgba;

	    /* Iterate through output buffer */
	    while(tar_ptr < tar_end)
	    {
		/* Solid? */
		if(src_ptr[3] >= 0x80)
		    *tar_ptr++ = (GifByteType)(
			((src_ptr[0] + src_ptr[1] + src_ptr[2]) / 3) *
			255 / 256
		    );
		else
		    *tar_ptr++ = (GifByteType)transparent_color_num;
		src_ptr += 4;
	    }
	}
	else
	{
	    GifByteType *tar_ptr = buf,
			*tar_end = tar_ptr + (width * height);
	    const u_int8_t *src_ptr = rgba;

	    /* Iterate through output buffer */
	    while(tar_ptr < tar_end)
	    {
		*tar_ptr++ = (GifByteType)(
		    (src_ptr[0] + src_ptr[1] + src_ptr[2]) / 3
		);
		src_ptr += 4;
	    }
	}

	/* At this point the GIF output buffer and the colormap are
	 * now ready to be written to the GIF file
	 */

	if(!(*user_aborted))
	{
	    /* Open GIF file for writing */
	    GifFileType *ft = EGifOpenFileName(filename, 0);
	    if(ft == NULL)
	    {
		FREE_ALL
		imgio_last_write_error = "Unable to open file for writing";
		return(-1);
	    }

	    /* Write screen descriptor */
	    EGifPutScreenDesc(
		ft,
		width, height,		/* Screen width & height */
		color_resolution,	/* Color resolution */
		bg_color_num,		/* Background color number */
		colormap		/* Screen colormap */
	    );

	    /* Write graphic control */
	    if(transparency && (transparent_color_num > -1))
	    {
		u_int8_t buf[] = {
		    0x01,       /* flags: transparency color flag on */
		    0x00,       /* delay time */
		    0x00,       /* delay time cntd */
		    (u_int8_t)transparent_color_num,
		};
		EGifPutExtension(
		    ft, GRAPHICS_EXT_FUNC_CODE,
		    sizeof(buf), buf
		);
	    }

	    /* Write image */	      
	    if(!(*user_aborted))
	    {
		/* Write image descriptor */
		EGifPutImageDesc(
		    ft,
		    0, 0,		/* Top & left */
		    width, height,	/* Width & height */
		    interlaced,		/* Interlace */
		    NULL		/* Local image colormap */
		);

		/* Write image data */
		if(interlaced)
		{
		    int i, y;

		    /* Interlace needs 4 passes on the image */
		    for(i = 0; i < 4; i++)
		    {
			for(y = gif_interlace_offset[i];
			    y < height;
			    y += gif_interlace_jumps[i]
			)
			{
			    EGifPutLine( 
				ft,
				(GifPixelType *)(buf + (y * width)),
				width
			    );
			}
			/* Report progress */
			if(progress_cb != NULL)
			{
			    if(!progress_cb(
				client_data,
				i, 4,
				width, height,
				bpl, bpp,
				rgba
			    ))
			    {
				*user_aborted = 1;
				break;
			    }
			}
		    }
		}    
		else
		{
		    int y;
		    GifByteType *buf_ptr = buf;
		    for(y = 0; y < height; y++)
		    {
		        EGifPutLine(ft, (GifPixelType *)buf_ptr, width);
		        buf_ptr += width;

		        /* Report progress */
		        if(progress_cb != NULL)
		        {
			    if(!progress_cb(
			        client_data,
			        y, height,
			        width, height,
			        bpl, bpp,
			        rgba
			    ))
			    {
			        *user_aborted = 1;
				break;
			    }
			}
		    }    
		}
	    }    

	    /* Write comments */ 
	    if(!STRISEMPTY(comments))
		EGifPutExtension(
		    ft, COMMENT_EXT_FUNC_CODE,
		    STRLEN(comments) + 1, comments
		);

	    /* Close GIF file */ 
	    EGifCloseFile(ft);
	}

	FREE_ALL

	if(*user_aborted)
	{
	    imgio_last_write_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(0);

#undef FREE_ALL
}

/*
 *	Writes the specified image data to a GIF file in black & white.
 */
static int ImgWriteGIFFileRGBAToBW(
	const char *filename,
	int width, int height,
	int bpl, int bpp,       /* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,       /* 4 bytes in RGBA format */
	const char *comments,
	int interlaced,         /* 0 or 1 */
	int transparency,       /* 0 or 1 */
	void *client_data,
	int (*progress_cb)(
		void *,         /* Client data */
		int, int,       /* Progress current & end values */
		int, int,       /* Width & height */
		int, int,       /* Bytes per line & bytes per pixel */
		u_int8_t *      /* RGBA data */
	),
	int *user_aborted
)
{
	int	color_resolution = 8,
		colormap_ncolors_set = 256,
		bg_color_num = 0,
		transparent_color_num = -1;
	ColorMapObject *colormap = NULL;
	GifByteType *buf = NULL;

#define FREE_ALL        {       \
 free(buf);                     \
 buf = NULL;                    \
				\
 if(colormap != NULL) {         \
  free(colormap->Colors);       \
  free(colormap);               \
  colormap = NULL;              \
 }                              \
}

	/* Create colormap */
	colormap = (ColorMapObject *)malloc(sizeof(ColorMapObject));
	colormap->ColorCount = colormap_ncolors_set;
	colormap->BitsPerPixel = 8;
	colormap->Colors = (GifColorType *)calloc(
	    colormap->ColorCount, sizeof(GifColorType)
	);
	/* Generate black and white colors */
	if(colormap->Colors != NULL)
	{
	    if(transparency)
	    {
		/* Generate 3 colors; black, white, and the rest
		 * are black (including color 255 for the transparent
		 * pixel)
		 */

		/* Set color 0 to black */
		int i = 0, m = colormap->ColorCount;
		GifColorType *c = &colormap->Colors[i];
		c->Red = c->Green = c->Blue = (GifByteType)0x00;

		/* Set color 1 to white */
		i = 1;
		if(i < m)
		{
		    c = &colormap->Colors[i]; 
		    c->Red = c->Green = c->Blue = (GifByteType)0xff;
		}

		/* Set color 2 to (transparent) to white */
		transparent_color_num = i = 2;
		if(i < m)
		{
		    c = &colormap->Colors[i];
		    c->Red = c->Green = c->Blue = (GifByteType)0x00;
		}

		/* Set rest of colors to black */
		for(i = 3; i < m; i++)
		{
		    c = &colormap->Colors[i];
		    c->Red = c->Green = c->Blue = (GifByteType)0x00;
		}
	    }
	    else
	    {
		/* Generate 3 colors; black, white, and the rest
		 * are black (including color 255 for the transparent
		 * pixel)
		 */
		   
		/* Set color 0 to black */
		int i = 0, m = colormap->ColorCount;
		GifColorType *c = &colormap->Colors[i];
		c->Red = c->Green = c->Blue = (GifByteType)0x00;

		/* Set color 1 to white */
		i = 1;
		if(i < m)
		{
		    c = &colormap->Colors[i];
		    c->Red = c->Green = c->Blue = (GifByteType)0xff;
		}

		/* Set rest of colors to black */
		for(i = 2; i < m; i++)
		{
		    c = &colormap->Colors[i];
		    c->Red = c->Green = c->Blue = (GifByteType)0x00;
		}
	    }
	}    

	/* Create buffer for rendering the output GIF image */
	buf = (GifByteType *)malloc(width * height * sizeof(GifByteType));
	if(buf == NULL)
	{
	    FREE_ALL
	    imgio_last_write_error = "Memory allocation error";
	    return(-3);
	}

	/* Copy/convert RGBA data to black & white output buffer */
	if(transparency && (transparent_color_num > -1))
	{
	    GifByteType *tar_ptr = buf,
			*tar_end = tar_ptr + (width * height);
	    const u_int8_t *src_ptr = rgba;

	    /* Iterate through output buffer */
	    while(tar_ptr < tar_end)
	    {
		/* Solid? */
		if(src_ptr[3] >= 0x80)
		    *tar_ptr++ = (GifByteType)(
			(((src_ptr[0] + src_ptr[1] + src_ptr[2]) / 3) >= 0x80) ?
			    1 : 0
		    );
		else
		    *tar_ptr++ = (GifByteType)transparent_color_num;
		src_ptr += 4;
	    }
	}
	else
	{
	    GifByteType *tar_ptr = buf,
			*tar_end = tar_ptr + (width * height);
	    const u_int8_t *src_ptr = rgba;

	    /* Iterate through output buffer */
	    while(tar_ptr < tar_end)  
	    {
		*tar_ptr++ = (GifByteType)(
		    (((src_ptr[0] + src_ptr[1] + src_ptr[2]) / 3) > 0x80) ?
			1 : 0
		);
		src_ptr += 4;
	    }
	}

	/* At this point the GIF output buffer and the colormap are
	 * now ready to be written to the GIF file
	 */

	if(!(*user_aborted))
	{
	    /* Open GIF file for writing */
	    GifFileType *ft = EGifOpenFileName(filename, 0);
	    if(ft == NULL)
	    {
		FREE_ALL
		imgio_last_write_error = "Unable to open file for writing";
		return(-1);
	    }

	    /* Write screen descriptor */
	    EGifPutScreenDesc(
		ft,
		width, height,		/* Screen width & height */
		color_resolution,	/* Color resolution */
		bg_color_num,		/* Background color number */
		colormap		/* Screen colormap */
	    );

	    /* Write graphic control */
	    if(transparency && (transparent_color_num > -1))
	    {
		u_int8_t buf[] = {
		    0x01,       /* flags: transparency color flag on */
		    0x00,       /* delay time */
		    0x00,       /* delay time cntd */
		    (u_int8_t)transparent_color_num,
		};
		EGifPutExtension(
		    ft, GRAPHICS_EXT_FUNC_CODE,
		    sizeof(buf), buf
		);
	    }

	    /* Write image */	      
	    if(!(*user_aborted))
	    {
		/* Write image descriptor */
		EGifPutImageDesc(
		    ft,
		    0, 0,		/* Top & left */
		    width, height,	/* Width & height */
		    interlaced,		/* Interlace */
		    NULL		/* Local image colormap */
		);

		/* Write image data */
		if(interlaced)
		{
		    int i, y;

		    /* Interlace needs 4 passes on the image */
		    for(i = 0; i < 4; i++)
		    {
			for(y = gif_interlace_offset[i];
			    y < height;
			    y += gif_interlace_jumps[i]
			)
			{
			    EGifPutLine(
				ft,     
				(GifPixelType *)(buf + (y * width)),
				width                               
			    );
			}     
			/* Report progress */
			if(progress_cb != NULL)
			{
			    if(!progress_cb(
				client_data,
				i, 4,
				width, height,
				bpl, bpp,
				rgba
			    ))
			    {
				*user_aborted = 1;
				break;
			    }
			}
		    }
		}    
		else
		{
		    int y;
		    GifByteType *buf_ptr = buf;
		    for(y = 0; y < height; y++)
		    {
		        EGifPutLine(ft, (GifPixelType *)buf_ptr, width);
		        buf_ptr += width;

		        /* Report progress */
		        if(progress_cb != NULL)
		        {
			    if(!progress_cb(
			        client_data,
			        y, height,
			        width, height,
			        bpl, bpp,   
			        rgba
			    ))
			    {
			        *user_aborted = 1;
			        break;
			    }
			}
		    }
		}
	    }

	    /* Write comments */ 
	    if(!STRISEMPTY(comments))
		EGifPutExtension(
		    ft, COMMENT_EXT_FUNC_CODE,
		    STRLEN(comments) + 1, comments
		);

	    /* Close GIF file */
	    EGifCloseFile(ft);
	}

	FREE_ALL

	if(*user_aborted)
	{
	    imgio_last_write_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(0);

#undef FREE_ALL
}

/*
 *	Writes the specified image data to a GIF file.
 *
 *	Return values:
 *
 *	0	Success
 *	-1	General error
 *	-2	Bad value
 *	-3	Systems error
 *	-4	User abort
 */
int ImgWriteGIFFileRGBA(
	const char *filename,
	int width, int height,
	int bpl, int bpp,	/* bpp must be 4 */
	u_int8_t *rgba,
	const u_int8_t *bg_color,	/* 4 bytes in RGBA format */
	const char *comments,
	int interlaced,		/* 0 or 1 */
	int format,		/* 0 = B&W
				 * 1 = Greyscale
				 * 2 = Color */
	int transparency,	/* 0 or 1 */
	void *client_data,
	int (*progress_cb)(
		void *,         /* Client data */
		int, int,       /* Progress current & end values */
		int, int,       /* Width & height */
		int, int,       /* Bytes per line & bytes per pixel */
		u_int8_t *      /* RGBA data */
	)
)
{
	int status;
	int user_aborted = 0;


	/* Reset global last write error message pointer */
	imgio_last_write_error = NULL;                     

	if(STRISEMPTY(filename) || (rgba == NULL) ||
	   (width <= 0) || (height <= 0) ||
	   (bpp != 4)
	)            
	{
	    imgio_last_write_error = "Invalid value used to describe image";
	    return(-1);
	}

	/* Automatically calculate bytes per line? */
	if(bpl <= 0)
	    bpl = width * bpp;

	/* Report initial progress */
	if(progress_cb != NULL)
	{
	    if(!progress_cb(
		client_data,
		0, height,
		width, height,
		bpl, bpp,
		rgba
	    ))
		user_aborted = 1;
	}

	/* Save by format */
	switch(format)
	{
	  case 2:
	    status = ImgWriteGIFFileRGBAToColor(
		filename,
		width, height, bpl, bpp,
		rgba,
		bg_color,
		comments,
		interlaced,
		transparency,
		client_data, progress_cb, &user_aborted
	    );
	    break;

	  case 1:
	    status = ImgWriteGIFFileRGBAToGreyscale(
		filename,
		width, height, bpl, bpp,
		rgba,
		bg_color,
		comments,
		interlaced,
		transparency,
		client_data, progress_cb, &user_aborted
	    );
	    break;

	  case 0:
	    status = ImgWriteGIFFileRGBAToBW(
		filename,
		width, height, bpl, bpp,
		rgba,
		bg_color,
		comments,
		interlaced,
		transparency,
		client_data, progress_cb, &user_aborted
	    );
	    break;

	  default:
	    imgio_last_write_error = "Unsupported GIF color format";
	    status = -2;
	    break;
	}


	/* Report final progress */
	if((progress_cb != NULL) && !user_aborted)
	{
	    if(!progress_cb(
		client_data,
		height, height,
		width, height,
		bpl, bpp,
		rgba
	    ))
		user_aborted = 1;              
	}                        

	if(user_aborted)
	{
	    imgio_last_write_error = "User aborted operation";
	    return(-4);
	}
	else
	    return(status);
}

#endif
